using elFinder.Net.AdvancedDemo.Models;
using elFinder.Net.AdvancedDemo.Services;
using elFinder.Net.AspNetCore.Extensions;
using elFinder.Net.Core.Extensions;
using elFinder.Net.Core.Plugins;
using elFinder.Net.Core.Services.Drawing;
using elFinder.Net.Drivers.FileSystem.Extensions;
using elFinder.Net.Drivers.FileSystem.Helpers;
using elFinder.Net.Drivers.FileSystem.Services;
using elFinder.Net.Plugins.FileSystemQuotaManagement.Extensions;
using elFinder.Net.Plugins.LoggingExample.Extensions;
using Microsoft.AspNetCore.Authentication.Cookies;
using Microsoft.AspNetCore.Authentication.JwtBearer;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.ResponseCompression;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using System.IO;

namespace elFinder.Net.AdvancedDemo
{
    public class Startup
    {
        public const string StorageFolder = ".storage";
        public const string TempFileFolder = ".temp";

        public Startup(IConfiguration configuration, IWebHostEnvironment env)
        {
            Configuration = configuration;
            WebRootPath = env.WebRootPath;
            StoragePath = new DirectoryInfo(StorageFolder).FullName;
            TempPath = new DirectoryInfo(TempFileFolder).FullName;
        }

        public static string WebRootPath { get; private set; }
        public static string StoragePath { get; private set; }
        public static string TempPath { get; private set; }

        public static string MapStoragePath(string path)
        {
            path = path.Replace("~/", "").TrimStart('/').Replace('/', '\\');
            return PathHelper.GetFullPath(PathHelper.SafelyCombine(StoragePath, StoragePath, path));
        }

        public IConfiguration Configuration { get; }

        // This method gets called by the runtime. Use this method to add services to the container.
        public void ConfigureServices(IServiceCollection services)
        {
            var pluginCollection = new PluginCollection();

            services.AddElFinderAspNetCore()
                .AddFileSystemDriver(typeof(ApplicationFileSystemDriver),
                    tempFileCleanerConfig: (opt) =>
                    {
                        opt.ScanFolders.Add(TempPath, TempFileCleanerOptions.DefaultUnmanagedLifeTime);
                    })
                .AddFileSystemQuotaManagement(pluginCollection,
                    fileSystemDriverImplType: typeof(ApplicationFileSystemDriver))
                .AddElFinderLoggingExample(pluginCollection)
                .AddElFinderPlugins(pluginCollection);

            // Custom IVideoEditor for generating video thumbnail
            services.AddSingleton<IVideoEditor, AppVideoEditor>();

            services.AddDbContext<DataContext>(options =>
                options.UseInMemoryDatabase(nameof(DataContext)));

            // Since elFinder use HTML5 download (hyper link), there is no Authorization header
            // --> better to use Cookies authentication (or mixed: need customization)
            services.AddAuthentication(JwtBearerDefaults.AuthenticationScheme)
                .AddJwtBearer(JwtBearerDefaults.AuthenticationScheme, options =>
                {
                    //...
                })
                .AddCookie(CookieAuthenticationDefaults.AuthenticationScheme, options =>
                {
                    //...
                });

            services.AddResponseCompression(options =>
            {
                options.EnableForHttps = true;
                options.Providers.Add<GzipCompressionProvider>();
            });

            services.Configure<GzipCompressionProviderOptions>(options => options.Level = System.IO.Compression.CompressionLevel.Fastest);

            services.AddRazorPages();
            services.AddControllersWithViews().AddNewtonsoftJson();
        }

        // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
        public void Configure(IApplicationBuilder app, IWebHostEnvironment env)
        {
            Directory.CreateDirectory(MapStoragePath("./upload"));

            if (env.IsDevelopment())
            {
                app.UseDeveloperExceptionPage();
            }
            else
            {
                app.UseExceptionHandler("/Error");
                // The default HSTS value is 30 days. You may want to change this for production scenarios, see https://aka.ms/aspnetcore-hsts.
                app.UseHsts();
            }

            app.UseHttpsRedirection();
            app.UseResponseCompression();
            app.UseStaticFiles();

            app.UseRouting();

            app.UseAuthentication();
            app.UseAuthorization();

            app.UseEndpoints(endpoints =>
            {
                endpoints.MapRazorPages();
                endpoints.MapControllers();
            });
        }
    }
}
